@@ -1,6 +1,4 @@
-from __future__ import annotations
-
-__doc__ = r"""
+r"""
 
 This module implements the SPARQL 1.1 Property path operators, as
 defined in:
@@ -182,8 +180,10 @@ No vars specified:
 
 """
 
+from __future__ import annotations
 
 import warnings
+from abc import ABC, abstractmethod
 from functools import total_ordering
 from typing import (
     TYPE_CHECKING,
@@ -214,29 +214,31 @@ ZeroOrOne = "?"
 
 
 def _n3(
-    arg: Union["URIRef", "Path"], namespace_manager: Optional["NamespaceManager"] = None
+    arg: Union[URIRef, Path], namespace_manager: Optional[NamespaceManager] = None
 ) -> str:
-    # type error: Item "Path" of "Union[Path, URIRef]" has no attribute "n3"  [union-attr]
     if isinstance(arg, (SequencePath, AlternativePath)) and len(arg.args) > 1:
         return "(%s)" % arg.n3(namespace_manager)
-    return arg.n3(namespace_manager)  # type: ignore[union-attr]
+    return arg.n3(namespace_manager)
 
 
 @total_ordering
-class Path:
-    __or__: Callable[["Path", Union["URIRef", "Path"]], "AlternativePath"]
-    __invert__: Callable[["Path"], "InvPath"]
-    __neg__: Callable[["Path"], "NegatedPath"]
-    __truediv__: Callable[["Path", Union["URIRef", "Path"]], "SequencePath"]
-    __mul__: Callable[["Path", str], "MulPath"]
+class Path(ABC):
+    __or__: Callable[[Path, Union[URIRef, Path]], AlternativePath]
+    __invert__: Callable[[Path], InvPath]
+    __neg__: Callable[[Path], NegatedPath]
+    __truediv__: Callable[[Path, Union[URIRef, Path]], SequencePath]
+    __mul__: Callable[[Path, str], MulPath]
 
+    @abstractmethod
     def eval(
         self,
-        graph: "Graph",
-        subj: Optional["_SubjectType"] = None,
-        obj: Optional["_ObjectType"] = None,
-    ) -> Iterator[Tuple["_SubjectType", "_ObjectType"]]:
-        raise NotImplementedError()
+        graph: Graph,
+        subj: Optional[_SubjectType] = None,
+        obj: Optional[_ObjectType] = None,
+    ) -> Iterator[Tuple[_SubjectType, _ObjectType]]: ...
+
+    @abstractmethod
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str: ...
 
     def __hash__(self):
         return hash(repr(self))
@@ -258,9 +260,9 @@ class InvPath(Path):
 
     def eval(
         self,
-        graph: "Graph",
-        subj: Optional["_SubjectType"] = None,
-        obj: Optional["_ObjectType"] = None,
+        graph: Graph,
+        subj: Optional[_SubjectType] = None,
+        obj: Optional[_ObjectType] = None,
     ) -> Generator[Tuple[_ObjectType, _SubjectType], None, None]:
         for s, o in eval_path(graph, (obj, self.arg, subj)):
             yield o, s
@@ -268,7 +270,7 @@ class InvPath(Path):
     def __repr__(self) -> str:
         return "Path(~%s)" % (self.arg,)
 
-    def n3(self, namespace_manager: Optional["NamespaceManager"] = None) -> str:
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
         return "^%s" % _n3(self.arg, namespace_manager)
 
 
@@ -283,9 +285,9 @@ class SequencePath(Path):
 
     def eval(
         self,
-        graph: "Graph",
-        subj: Optional["_SubjectType"] = None,
-        obj: Optional["_ObjectType"] = None,
+        graph: Graph,
+        subj: Optional[_SubjectType] = None,
+        obj: Optional[_ObjectType] = None,
     ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
         def _eval_seq(
             paths: List[Union[Path, URIRef]],
@@ -325,7 +327,7 @@ class SequencePath(Path):
     def __repr__(self) -> str:
         return "Path(%s)" % " / ".join(str(x) for x in self.args)
 
-    def n3(self, namespace_manager: Optional["NamespaceManager"] = None) -> str:
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
         return "/".join(_n3(a, namespace_manager) for a in self.args)
 
 
@@ -340,9 +342,9 @@ class AlternativePath(Path):
 
     def eval(
         self,
-        graph: "Graph",
-        subj: Optional["_SubjectType"] = None,
-        obj: Optional["_ObjectType"] = None,
+        graph: Graph,
+        subj: Optional[_SubjectType] = None,
+        obj: Optional[_ObjectType] = None,
     ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
         for x in self.args:
             for y in eval_path(graph, (subj, x, obj)):
@@ -351,7 +353,7 @@ class AlternativePath(Path):
     def __repr__(self) -> str:
         return "Path(%s)" % " | ".join(str(x) for x in self.args)
 
-    def n3(self, namespace_manager: Optional["NamespaceManager"] = None) -> str:
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
         return "|".join(_n3(a, namespace_manager) for a in self.args)
 
 
@@ -374,9 +376,9 @@ class MulPath(Path):
 
     def eval(
         self,
-        graph: "Graph",
-        subj: Optional["_SubjectType"] = None,
-        obj: Optional["_ObjectType"] = None,
+        graph: Graph,
+        subj: Optional[_SubjectType] = None,
+        obj: Optional[_ObjectType] = None,
         first: bool = True,
     ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
         if self.zero and first:
@@ -475,7 +477,7 @@ class MulPath(Path):
     def __repr__(self) -> str:
         return "Path(%s%s)" % (self.path, self.mod)
 
-    def n3(self, namespace_manager: Optional["NamespaceManager"] = None) -> str:
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
         return "%s%s" % (_n3(self.path, namespace_manager), self.mod)
 
 
@@ -509,7 +511,7 @@ class NegatedPath(Path):
     def __repr__(self) -> str:
         return "Path(! %s)" % ",".join(str(x) for x in self.args)
 
-    def n3(self, namespace_manager: Optional["NamespaceManager"] = None) -> str:
+    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
         return "!(%s)" % ("|".join(_n3(arg, namespace_manager) for arg in self.args))
 
 
@@ -538,9 +540,9 @@ def path_sequence(self: Union[URIRef, Path], other: Union[URIRef, Path]):
 def evalPath(  # noqa: N802
     graph: Graph,
     t: Tuple[
-        Optional["_SubjectType"],
+        Optional[_SubjectType],
         Union[None, Path, _PredicateType],
-        Optional["_ObjectType"],
+        Optional[_ObjectType],
     ],
 ) -> Iterator[Tuple[_SubjectType, _ObjectType]]:
     warnings.warn(
@@ -556,9 +558,9 @@ def evalPath(  # noqa: N802
 def eval_path(
     graph: Graph,
     t: Tuple[
-        Optional["_SubjectType"],
+        Optional[_SubjectType],
         Union[None, Path, _PredicateType],
-        Optional["_ObjectType"],
+        Optional[_ObjectType],
     ],
 ) -> Iterator[Tuple[_SubjectType, _ObjectType]]:
     return ((s, o) for s, p, o in graph.triples(t))
